#ifndef INC_ARBOL_BINARIO
#define INC_ARBOL_BINARIO
#include <exception>
#include <cstdlib>
   using namespace std;

template <class T>
class ArbolBinario {
   public:
   // Constructoras
      /**
       * Representa el rbol vaco
       */
      ArbolBinario();
   
      /**
       * Crea un nuevo rbol a partir de un elemento y dos subrboles
       * con semntica de alias
       */
      ArbolBinario(const T &,const ArbolBinario<T> &,
            const ArbolBinario<T> &);
   
      /**
       * Modifica un rbol poniendo 'e' como raz e izqdo y drcho
       * como subrboles
       * (si el rbol est vaco, se produce la excepcin ArbolVacio)
       */
      void modificar(const T &e, const ArbolBinario<T> &izqdo,
        const ArbolBinario<T> &drcho);

      /**
       * Modifica el rbol actual para representar un rbol vaco. 
       * El valor anterior es destruido de forma controlada.
       */
      void vaciar();
      
      /**
       * Devuelve el subrbol izquierdo del rbol.
       * Si est vaco lanza la excepcin ArbolVacio
       */
      ArbolBinario<T> izquierdo() const;
   
      /**
       * Devuelve el subrbol derecho del rbol.
       * Si est vaco lanza la excepcin ArbolVacio
       */
      ArbolBinario<T> derecho() const; 
   
      /**
       * Realiza una copia del rbol binario que representa la variable
       */
      ArbolBinario<T> copia();
      
   // Observadoras
      /**
       * Devuelve la informacin asociada a la raz del rbol
       */
      T raiz() const;
   
      /**
       * Devuelve verdadero si el nmero de elementos es distinto de cero
       */
      bool hayElementos() const;
   
      /**
       * Se lanza cuando se intenta realizar una operacin
       * no aplicable a los rboles vacos
       */
      class ArbolVacio: public exception{};
   private:
      /**
       * El tipo Nodo es privado, ya que su razn de ser es servir de 
       * "ladrillo" para formar la estructura del rbol con un campo
       * de informacin y dos campos de encadenamiento
       */
      struct Nodo {
         T elemento;
         ArbolBinario<T> hijos[2];
         Nodo(T e, ArbolBinario<T> izq, ArbolBinario<T> der) {
            elemento = e;
            hijos[0] = izq;
            hijos[1] = der;
         }
         ~Nodo(){}
      };
      typedef Nodo *PNodo;
      PNodo arbol;  // Puntero a la raz del rbol
};

/**
 * Representa el rbol vaco
 */
template <class T>
ArbolBinario<T>::ArbolBinario() {
   arbol = NULL;
};

/**
 * Crea un nuevo rbol a partir de un elemento y dos subrboles
 * con semntica de alias
 */
template <class T>
ArbolBinario<T>::ArbolBinario(const T &e,const ArbolBinario<T> &izqdo,
      const ArbolBinario<T> &drcho) {
   arbol = new Nodo(e,izqdo,drcho);
};

/**
 * Modifica un rbol poniendo 'e' como raz e izqdo y drcho
 * como subrboles
 * (si el rbol est vaco, se produce la excepcin ArbolVacio)
 */
template <class T>
void ArbolBinario<T>::modificar(const T &e, const ArbolBinario<T> &izqdo,
        const ArbolBinario<T> &drcho) {
   if (!hayElementos()) throw ArbolVacio();
   
   arbol->elemento = e;
   arbol->hijos[0] = izqdo;
   arbol->hijos[1] = drcho;
}

/**
 * Modifica el rbol actual para representar un rbol vaco. 
 * El valor anterior es destruido de forma controlada.
 */
template <class T>
void ArbolBinario<T>::vaciar() {
    if (hayElementos()) {
        arbol->hijos[0].vaciar();
        arbol->hijos[1].vaciar();
        delete arbol;
        arbol = NULL;
    }
}

/**
 * Devuelve el subrbol izquierdo del rbol.
 * Si est vaco lanza la excepcin ArbolVacio
 */
template <class T>
ArbolBinario<T> ArbolBinario<T>::izquierdo() const { 
   if (!hayElementos()) throw ArbolVacio();
   
   return arbol->hijos[0];
}; 

/**
 * Devuelve el subrbol derecho del rbol.
 * Si est vaco lanza la excepcin ArbolVacio
 */
template <class T>
ArbolBinario<T> ArbolBinario<T>::derecho() const { 
   if (!hayElementos()) throw ArbolVacio();
   
   return arbol->hijos[1];
}; 

/**
 * Realiza una copia del rbol binario que representa la variable
 */
template<class T>
ArbolBinario<T> ArbolBinario<T>::copia()  {         
   if (!hayElementos()) {
      return ArbolBinario<T>();
   }
   else {
      return ArbolBinario<T>(arbol->elemento,izquierdo().copia(),
                        derecho().copia());
   }
};

/**
 * Devuelve la informacin asociada a la raz del rbol
 */
template <class T>
T ArbolBinario<T>::raiz() const {
   return arbol->elemento;
};

/**
 * Devuelve verdadero si el nmero de elementos es distinto de cero
 */
template <class T>
bool ArbolBinario<T>::hayElementos() const {
   return arbol != NULL;
};

#endif
